#define IMPORT_CORE
#include "../../src/ErrorSimulator.h"
#include "../Tests.h"
#include "core/Utility.h"

static const float eps = 0.0001f;

static void testPopCount() {
    TEST_U64(4, popCount(0xF));
    TEST_U64(0, popCount(0x0));
    TEST_U64(2, popCount(0x6));
    TEST_U64(7, popCount(0x7F));
    TEST_U64(3, popCount(0x2A));
    TEST_U64(32, popCount(0xFFFFFFFF));
    TEST_U64(64, popCount(0xFFFFFFFFFFFFFFFF));
    TEST_U64(44, popCount(0xFFFF0FFFFFFF));
}

static void testZeroRellocate() {
    void* buffer = cReallocate(nullptr, 16);
    TEST_NOT_NULL(buffer);
    buffer = cReallocate(buffer, 0);
    TEST_NULL(buffer);
}

static void testMemoryInfoList() {
    void* a = cAllocate(8);
    void* b = cAllocate(8);
    void* c = cAllocate(8);
    void* d = cAllocate(8);
    cFree(b); // remove middle element
    cFree(a); // remove first
    cFree(d); // remove last
    cFree(c); // remove single
    cFree(nullptr);
}

static void testInterpolate() {
    TEST_FLOAT(7.5f, interpolate(5.0f, 10.0f, 0.5f), eps);
    TEST_FLOAT(-2.0, interpolate(-10.0f, 10.0f, 0.4f), eps);
    TEST_FLOAT(10.0f, interpolate(-3.0f, 10.0f, 1.0f), eps);
    TEST_FLOAT(7.0f, interpolate(7.0f, 10.0f, 0.0f), eps);
    TEST_FLOAT(6.0f, interpolate(0.0f, 10.0f, 0.6f), eps);
}

static void testRadianToDegree() {
    TEST_FLOAT(45.0f, radianToDegree(PI * 0.25f), eps);
    TEST_FLOAT(90.0f, radianToDegree(PI * 0.5f), eps);
    TEST_FLOAT(180.0f, radianToDegree(PI), eps);
    TEST_FLOAT(360.0f, radianToDegree(PI * 2.0f), eps);
}

static void testDegreeToRadian() {
    TEST_FLOAT(PI * 0.25f, degreeToRadian(45.0f), eps);
    TEST_FLOAT(PI * 0.5f, degreeToRadian(90.0f), eps);
    TEST_FLOAT(PI, degreeToRadian(180.0f), eps);
    TEST_FLOAT(PI * 2.0f, degreeToRadian(360.0f), eps);
}

static void testSleep(i64 nanos) {
    i64 time = -getNanos();
    TEST_FALSE(sleepNanos(nanos));
    time += getNanos();
    TEST_TRUE(time >= nanos && time <= (nanos * 13) / 10);
}

typedef struct {
    int i;
    i64 d;
} SwapTest;

static void testSwap() {
    SwapTest a = {3, 20};
    SwapTest b = {7, 30};
    swap(&a, &b);
    TEST_INT(7, a.i);
    TEST_INT(3, b.i);
    TEST_I64(30, a.d);
    TEST_I64(20, b.d);
}

static void testFail() {
#ifdef ERROR_SIMULATOR
    failTimeGet = true;
    TEST_I64(-1, getNanos());
    failTimeGet = false;
#endif
}

void testUtility(bool light) {
    testPopCount();
    testZeroRellocate();
    testMemoryInfoList();
    testInterpolate();
    testRadianToDegree();
    testDegreeToRadian();
    testSleep(light ? 5'000'000 : 50'000'000);
    testSleep(light ? 50'000'000 : 1'300'000'000);
    testSwap();
    testFail();
}

static void outOfMemory(void*) {
    setOutOfMemoryHandler(nullptr, nullptr);
}

void testInvalidAllocate(void) {
    setOutOfMemoryHandler(outOfMemory, nullptr);
    cAllocate(0xFFFFFFFFFFF);
    TEST_TRUE(false);
    finalizeTests();
    EXIT(0);
}

void testInvalidReallocate(void) {
    setOutOfMemoryHandler(outOfMemory, nullptr);
    void* p = cAllocate(0xFF);
    printMemoryReport();
    cReallocate(p, 0xFFFFFFFFFFF);
    TEST_TRUE(false);
    finalizeTests();
    EXIT(0);
}

[[noreturn]] void testPreCanary(void) {
#ifdef CORE_CHECK_MEMORY
    char* p = cAllocate(16);
    p[-1] = 0;
    cFree(p);
    TEST_TRUE(false);
#endif
    finalizeTests();
    EXIT(0);
}

[[noreturn]] void testPostCanary(void) {
#ifdef CORE_CHECK_MEMORY
    char* p = cAllocate(16);
    p[17] = 0;
    cFree(p);
    TEST_TRUE(false);
#endif
    finalizeTests();
    EXIT(0);
}
